package de.hft.stuttgart.volumecalculator;

/**
 * A volume provider provides the volume of a subject and its bounding box. A subject can be 
 * single geometry, a building part or a whole building. Depending on the applied calculation
 * method for the bounding boxes, they can be axis-aligned or object-oriented bounding boxes of
 * different accuracy.
 * 
 * If the volume of the bounding box exceeds the volume of the subject, then the subject volume
 * is wrong and the volume calculation of the subject failed. This is only true under the
 * assumption, that the calculation of the bounding box is correct. You can test for detected errors
 * with the method calculationErrorDetected(). The method getCheckedVolume() will return the
 * subject volume, if no error will be detected. If an error is detected, getCheckedVolume() returns
 * the volume of the bounding box as a fall-back solution. 
 * 
 * If the subject is a building, then its volume includes the volumes of its building parts. Note
 * that City Doctor forbids building parts to consist of further building parts.  
 * 
 * @author Marcel Bruse
 */
public class VolumeProvider {

	/**
	 * This volume is either the volume of the subject or the volume of the subject's bounding box.
	 * If the volume of the subject exceeds the volume of it's bounding box, then checked volume will
	 * be the volume of the bounding box. Otherwise, checked volume will be the volume of the subject.
	 */
	private double checkedVolume;
	
	/** The calculated volume of the subject. */
	private double subjectVolume;
	
	/** The calculated volume of the bounding box. */
	private double boundingBoxVolume;
	
	/**
	 * This flag indicates whether an error has been detected in the calculated volume of the
	 * subject. If it is true, then an error has been detected.
	 */
	private boolean errorDetected;
	
	/**
	 * This constructor forces the provider to maintain the volumes of the subject and it's
	 * bounding box.
	 * 
	 * @param subjectVolume The volume of the subject.
	 * @param boundingBoxVolume The volume of the subject's bounding box.
	 */
	public VolumeProvider(double subjectVolume, double boundingBoxVolume) {
		this.subjectVolume = subjectVolume;
		this.boundingBoxVolume = boundingBoxVolume;
		if (subjectVolume < 0.0 || subjectVolume > boundingBoxVolume) {
			checkedVolume = boundingBoxVolume;
			errorDetected = true;
		} else {
			checkedVolume = subjectVolume;
			errorDetected = false;
		}
	}

	/**
	 * @return The volume of the subject.
	 */
	public double getSubjectVolume() {
		return subjectVolume;
	}
	
	/**
	 * @return The volume of the bounding box.
	 */
	public double getBoundingBoxVolume() {
		return boundingBoxVolume;
	}
	
	/**
	 * @return The returned volume is either the volume of the subject or the volume of the subject's
	 * bounding box. If the volume of the subject exceeds the volume of it's bounding box, then a 
	 * calculation error has been detected and checked volume will be the volume of the bounding box.
	 * Otherwise, checked volume will be the volume of the subject.
	 * 
	 * Ergo, if the checked volume is the building volume, this doesn't provide that the calculated
	 * volume is correct. There may still be a error, but is has not been detected.
	 */
	public double getCheckedVolume() {
		return checkedVolume;
	}
	
	/**
	 * @return Returns true, if the volume of the subject exceeds the volume of it's bounding box or is less than 0.
	 */
	public boolean calculationErrorDetected() {
		return errorDetected;
	}
	
}
