package eu.simstadt.ade.energy;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import de.hft.stuttgart.citydoctor.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor.datastructure.Building;
import de.hft.stuttgart.citydoctor.datastructure.Extension;

/**
 * This extension represents one or more components of the Energy ADE and extends the CityGML feature named
 * AbstractBoundarySurface and all its subclasses.
 *  
 * @author Marcel Bruse
 */
public class BoundarySurfaceUnit extends Extension {

	/** The namespace URI of the Energy ADE */
	public static final String NAMESPACE_URI = "http://simstadt.eu/ade/energy";
	
	/** The local name of the corresponding XML element. */
	public static final String LOCAL_NAME = "";
	
	/** 
	 * The qualified name of this particular ADE element. It consists of the namespace of the ADE and the local name
	 * of the XML element.
	 */
	public static final String QUALIFIED_NAME = NAMESPACE_URI + "/" + LOCAL_NAME;
	
	/** A list of the neighboring buildings of the parent building. */
	private List<Building> adjacentBuildings;
	
	/** The exterior surface area. TODO: Explain this field. What does it mean? Area of what? */
	private Double exteriorSurfaceArea;
	
	/** The shared surface area. This is the area of this surface that bounds at least two buildings. */
	private Double sharedSurfaceArea;
	
	/** The thermal boundary area above ground. TODO: Explain this field. */
	private Double thermalBoundaryAreaAboveGround;
	
	/** The thermal boundary area below ground. TODO: Explain this field. */
	private Double thermalBoundaryAreaBelowGround;
	
	/** The azimuth of this boundary surface. */
	private Integer azimuth;
	
	/** The inclination of this boundary surface. */
	private Integer inclination;
	
	/** List of short wave irradiance values. TODO: Explain this field. */ 
	private List<Integer> shortWaveIrradiance;
	
	/** The shell properties of this boundary surface. */
	private ShellProperties shellProperties;
	
	/** The standard constructor */
	public BoundarySurfaceUnit() {
		adjacentBuildings = new ArrayList<Building>();
	}
	
	/** A constructor for the boundary surface units.
	 * @param boundarySurface The parent boundary surface that is enriched by this extension.
	 */
	public BoundarySurfaceUnit(BoundarySurface boundarySurface) {
		this();
		setParent(boundarySurface);
	}
	
	/**
	 * @return the type of the parent boundary surface.
	 */
	public BoundarySurface.boundarySurfaceType getBoundarySurfaceType() {
		return getBoundarySurface().getType();
	}
	
	/**
	 * @return the parent boundary surface.
	 */
	public BoundarySurface getBoundarySurface() {
		return (BoundarySurface) getParent();
	}
	
	/**
	 * @return the list of all neighboring buildings of the parent building. 
	 */
	public List<Building> getAdjacentBuildings() {
		return adjacentBuildings;
	}

	/**
	 * Maintains the list of all neighboring buildings of the parent building.
	 * 
	 * @param adjacentBuildings the list of all neighboring buildings of the parent buliding.
	 */
	public void setAdjacentBuildings(List<Building> adjacentBuildings) {
		this.adjacentBuildings = adjacentBuildings;
	}
	
	/**
	 * @return the building unit associated with the same city doctor building as this boundary surface unit.
	 */
	public BuildingUnit getBuildingUnit() {
		return (BuildingUnit) getBoundarySurface().getParent().getExtensions().get(NAMESPACE_URI);
	}

	/**
	 * @return the exterior surface area of the parent boundary surface.
	 */
	public Double getExteriorSurfaceArea() {
		return exteriorSurfaceArea;
	}

	/**
	 * Sets the exterior surface area of the parent boundary surface.
	 * 
	 * @param exteriorSurfaceArea The exterior surface area of the parent boundary surface.
	 */
	public void setExteriorSurfaceArea(Double exteriorSurfaceArea) {
		this.exteriorSurfaceArea = exteriorSurfaceArea;
	}
	
	/**
	 * @return True, if the exterior surface area is not null. False, otherwise.
	 */
	public boolean isSetExteriorSurfaceArea() {
		return Objects.nonNull(exteriorSurfaceArea);
	}

	/**
	 * @return the shared surface area of the parent boundary surface.
	 */
	public Double getSharedSurfaceArea() {
		return sharedSurfaceArea;
	}

	/**
	 * Sets the shared surface area of the parent boundary surface.
	 * 
	 * @param sharedSurfaceArea The blind surface area of the parent boundary surface.
	 */
	public void setSharedSurfaceArea(Double sharedSurfaceArea) {
		this.sharedSurfaceArea = sharedSurfaceArea;
	}
	
	/**
	 * @return True, if the shared surface area is not null. False, otherwise.
	 */
	public boolean isSetSharedSurfaceArea() {
		return Objects.nonNull(sharedSurfaceArea);
	}

	/**
	 * @return the thermal surface area above ground of the parent boundary surface.
	 */
	public Double getThermalBoundaryAreaAboveGround() {
		return thermalBoundaryAreaAboveGround;
	}

	/**
	 * Sets the thermal surface area above ground of the parent boundary surface.
	 * 
	 * @param thermalBoundaryAreaAboveGround The thermal surface area above ground of the parent boundary surface.
	 */
	public void setThermalBoundaryAreaAboveGround(Double thermalBoundaryAreaAboveGround) {
		this.thermalBoundaryAreaAboveGround = thermalBoundaryAreaAboveGround;
	}

	/**
	 * @return True, if the thermal boundary area above ground is not null. False, otherwise.
	 */
	public boolean isSetThermalBoundaryAreaAboveGround() {
		return Objects.nonNull(thermalBoundaryAreaAboveGround);
	}

	/**
	 * @return the thermal surface area below ground of the parent boundary surface.
	 */
	public Double getThermalBoundaryAreaBelowGround() {
		return thermalBoundaryAreaBelowGround;
	}

	/**
	 * Sets the thermal surface area below ground of the parent boundary surface.
	 * 
	 * @param thermalBoundaryAreaBelowGround The thermal surface area below ground of the parent boundary surface.
	 */
	public void setThermalBoundaryAreaBelowGround(Double thermalBoundaryAreaBelowGround) {
		this.thermalBoundaryAreaBelowGround = thermalBoundaryAreaBelowGround;
	}
	
	/**
	 * @return True, if the thermal boundary area below ground is not null. False, otherwise.
	 */
	public boolean isSetThermalBoundaryAreaBelowGround() {
		return Objects.nonNull(thermalBoundaryAreaBelowGround);
	}
	
	/**
	 * @return the azimuth of the the parent boundary surface.
	 */
	public Integer getAzimuth() {
		return azimuth;
	}

	/**
	 * Sets the azimuth of the parent boundary surface.
	 * 
	 * @param azimuth The azimuth of the parent boundary surface.
	 */
	public void setAzimuth(Integer azimuth) {
		this.azimuth = azimuth;
	}
	
	/**
	 * @return True, if the azimuth is not null. False, otherwise. 
	 */
	public boolean isSetAzimuth() {
		return Objects.nonNull(azimuth);
	}

	/**
	 * @return the inclination of the parent boundary surface.
	 */
	public Integer getInclination() {
		return inclination;
	}

	/**
	 * Sets the inclination of the parent boundary surface.
	 * 
	 * @param inclination The inclination of the parent boundary surface.
	 */
	public void setInclination(Integer inclination) {
		this.inclination = inclination;
	}
	
	/**
	 * @return True, if the inclination is not null. False, otherwise.
	 */
	public boolean isSetInclination() {
		return Objects.nonNull(inclination);
	}

	/**
	 * @return the short wave irradiance of the parent boundary surface.
	 */
	public List<Integer> getShortWaveIrradiance() {
		return shortWaveIrradiance;
	}

	/**
	 * Sets the short wave irradiance of the parent boundary surface.
	 * 
	 * @param shortWaveIrradiance the short wave irradiance of the parent boundary surface.
	 */
	public void setShortWaveIrradiance(List<Integer> shortWaveIrradiance) {
		this.shortWaveIrradiance = shortWaveIrradiance;
	}
	
	/**
	 * @return True, if the inclination is not null. False, otherwise.
	 */
	public boolean isSetShortWaveIrradiance() {
		return Objects.nonNull(shortWaveIrradiance);
	}

	/**
	 * @return the shell properties of the parent boundary surface.
	 */
	public ShellProperties getShellProperties() {
		return shellProperties;
	}

	/**
	 * Sets the shell properties of the parent boundary surface.
	 * 
	 * @param shellProperties the shell properties of the parent boundary surface.
	 */
	public void setShellProperties(ShellProperties shellProperties) {
		this.shellProperties = shellProperties;
	}
	
	/**
	 * @return True, if the shell properties are set. False, otherwise.
	 */
	public boolean isSetShellProperties() {
		return Objects.nonNull(shellProperties);
	}
	
	/**
	 * @return the namespace URI of the Energy ADE.
	 */
	@Override
	public String getNamespaceURI() {
		return NAMESPACE_URI;
	}

	/**
	 * @return The local name of the corresponding XML element.
	 */
	@Override
	public String getLocalName() {
		return LOCAL_NAME;
	}

	/**
	 * @return Returns a copy of this object and all of its associated (child) objects.
	 */
	@Override
	public Extension deepCopy() {
		// TODO Auto-generated method stub
		return null;
	}
	
}
