package eu.simstadt.ade.energy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

import de.hft.stuttgart.citydoctor.datastructure.AbstractBuilding;
import de.hft.stuttgart.citydoctor.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor.datastructure.Building;
import de.hft.stuttgart.citydoctor.datastructure.Extension;
import de.hft.stuttgart.citydoctor.datastructure.Geometry.LOD;

/**
 * This extension represents one or more components of the Energy ADE and extends the CityGML feature named Building.
 *  
 * @author Marcel Bruse
 */
public class BuildingUnit extends Extension {

	/** The namespace URI of the Energy ADE */
	public static final String NAMESPACE_URI = "http://simstadt.eu/ade/energy";
	
	/** The local name of the corresponding XML element. */
	public static final String LOCAL_NAME = "";
	
	/** The qualified name of the corresponding XML element. */
	public static final String QUALIFIED_NAME = NAMESPACE_URI + "/" + LOCAL_NAME;
	
	/** The level of detail for which all other values have been calculated. */
	private LOD lod;
	
	/** A reference to the physics properties object defined by the Energy ADE */
	private PhysicsProperties physicsProperties;
	
	/** A reference to the spatial properties object defined by the Energy ADE */
	private SpatialProperties spatialProperties;
	
	/** A reference to the energy usages defined by the Energy ADE */
	private HashMap<EndUseType, EndUse> endUseMap = new HashMap<>();
	
	/** A reference to the list of usage zones associated with the parent building. */
	private UsageZoneList usageZoneList;
	
	/**
	 * The standard constructor.
	 */
	public BuildingUnit() {}
	
	/**
	 * A constructor for this building unit.
	 * 
	 * @param abstractBuilding The parent building, that is extended by this extension.
	 */
	public BuildingUnit(AbstractBuilding abstractBuilding) {
		this();
		setParent(abstractBuilding);
	}
	
	/**
	 * @return Returns the parent building of this application domain extension.
	 */
	public AbstractBuilding getAbstractBuilding() {
		return (AbstractBuilding) getParent();
	}
	
	/**
	 * @return all the boundary surface units associated with the parent building.
	 */
	public ArrayList<BoundarySurfaceUnit> getBoundarySurfaceUnits() {
		ArrayList<BoundarySurfaceUnit> boundarySurfaceUnits = new ArrayList<>();
		for (BoundarySurface boundarySurface : getAbstractBuilding().getBoundarySurfaceList()) {
			for (Extension extension : boundarySurface.getExtensions(NAMESPACE_URI)) {
				boundarySurfaceUnits.add((BoundarySurfaceUnit) extension);					
			}
		}
		return boundarySurfaceUnits;
	}
	
	/**
	 * @return the list of all adjacent buildings of the parent building.
	 */
	public Set<Building> adjacentBuildings() {
		Set<Building> adjacentBuildings = new HashSet<>();
		for (BoundarySurfaceUnit bsu : getBoundarySurfaceUnits()) {
			adjacentBuildings.addAll(bsu.getAdjacentBuildings());
		}
		return Collections.unmodifiableSet(adjacentBuildings);
	}
	
	/**
	 * @return the physics properties of the enriched building.
	 */
	public PhysicsProperties getPhysicsProperties() {
		return physicsProperties;
	}

	/**
	 * Sets the physics properties of the enriched building.
	 * 
	 * @param physicsProperties The building physics of the Energy ADE.
	 */
	public void setPhysicsProperties(PhysicsProperties physicsProperties) {
		this.physicsProperties = physicsProperties;
	}

	/**
	 * @return the spatial properties of the enriched building. 
	 */
	public SpatialProperties getSpatialProperties() {
		return spatialProperties;
	}

	/**
	 * Sets the spatial properties of the enriched building.
	 * 
	 * @param spatialProperties The building dimensions of the Energy ADE.
	 */
	public void setSpatialProperties(SpatialProperties spatialProperties) {
		this.spatialProperties = spatialProperties;
	}

	/**
	 * Adds a new end use object to the map of end usages.
	 * 
	 * @param endUseType The end use type of the end use object.
	 * @param endUse The end use object to be added.
	 */
	public void addEndUse(EndUseType endUseType, EndUse endUse) {
		endUseMap.put(endUseType, endUse);
	}
	
	/**
	 * @param endUseType The end use type of the end use object to be returned.
	 * @return Returns the end use object for the given end use type, if it exists.
	 */
	public EndUse getEndUse(EndUseType endUseType) {
		return endUseMap.get(endUseType);
	}
	
	/**
	 * @return Returns the whole end use map.
	 */
	public HashMap<EndUseType, EndUse> getEndUseMap() {
		return endUseMap;
	}
	
	/**
	 * Removes an existing end use object for a given end use type.
	 * @param endUseType The end use type of the end use object to be removed.
	 */
	public void removeEndUse(EndUseType endUseType) {
		endUseMap.remove(endUseType);
	}

	/**
	 * @return the list of all defined usage zones associated with the parent building.
	 */
	public UsageZoneList getUsageZoneList() {
		return usageZoneList;
	}

	/**
	 * Sets the list off all defined usage zones associaged with the parent building.
	 * 
	 * @param usageZoneList the list of all defined usage zones associated with the parent building.
	 */
	public void setUsageZoneList(UsageZoneList usageZoneList) {
		this.usageZoneList = usageZoneList;
	}

	/**
	 * @return The level of detail for which all other values have been calculated.
	 */
	public LOD getLOD() {
		return lod;
	}

	/**
	 * Sets the level of detail for which all other values will be / have been calculated.
	 * 
	 * @param lod The level of detail to be set.
	 */
	public void setLOD(LOD lod) {
		this.lod = lod;
	}
	
	/**
	 * @return True, if LOD is set. False, if it is null.
	 */
	public boolean isSetLOD() {
		return Objects.nonNull(lod);
	}
	
	/**
	 * @return The namespace URI of the Energy ADE.
	 */
	@Override
	public String getNamespaceURI() {
		return NAMESPACE_URI;
	}

	/**
	 * @return The local name of the corresponding XML element.
	 */
	@Override
	public String getLocalName() {
		return LOCAL_NAME;
	}

	/**
	 * Creates a deep copy of this building unit instance. This means that it will return also exact copies
	 * of all the objects which are associated with this building unit.
	 * 
	 * @return This method returns an exact copy of this building unit and all its associated objects. 
	 */
	@Override
	public Extension deepCopy() {
		// TODO Auto-generated method stub
		return null;
	}
	
}
