package eu.simstadt.ade.energy;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

import org.citygml4j.builder.jaxb.marshal.JAXBMarshaller;
import org.citygml4j.model.citygml.ade.ADEComponent;
import org.citygml4j.model.citygml.building.Building;
import org.citygml4j.model.citygml.building.BuildingPart;
import org.citygml4j.model.citygml.building.ClosureSurface;
import org.citygml4j.model.citygml.building.GroundSurface;
import org.citygml4j.model.citygml.building.OuterCeilingSurface;
import org.citygml4j.model.citygml.building.OuterFloorSurface;
import org.citygml4j.model.citygml.building.RoofSurface;
import org.citygml4j.model.citygml.building.WallSurface;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import de.hft.stuttgart.citydoctor.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor.datastructure.Extension;
import de.hft.stuttgart.citydoctor.datastructure.Geometry.LOD;
import de.hft.stuttgart.citydoctor.mapper.ExtensionHandler;

/**
 * This ExtensionHandler is able to marshal and unmarshal elements of the Energy ADE.
 * Marshalling means that instances of the classes of the package eu.simstadt.ade.energy
 * will be serialized to instances of w3c.dom.Elements and wrapped in ADEComponent objects. These
 * classes implement City Doctor's abstract class Extension. Unmarshalling inverts the marshalling operation.
 * 
 * This ExtensionHandler will be found by City Doctor's service loader and registered at City Doctor's
 * ExtensionRegistry as soon as you load a CityGML file.
 * 
 * @author Marcel Bruse
 */
public class EnergyADEHandler extends ExtensionHandler {

	/** The package of the Energy ADE. */
	private final String ADE_PACKAGE = "eu.simstadt.ade.energy";
	
	/** The namespace URI of the Energy ADE. */
	private final String NAMESPACE_URI = "http://simstadt.eu/ade/energy";
	
	/** The local name of the building physics tag. */
	private final String PHYSICS_PROPERTIES = "physicsProperties";
	
	/** The local name of the building dimensions tag. */
	private final String SPATIAL_PROPERTIES = "spatialProperties";
	
	/** The local name of the energy demand tag. */
	private final String END_USE = "endUse";
	
	/** The local name of the usage zone list tag. */
	private final String USAGE_ZONE_LIST = "usageZoneList";
	
	/** The devel of detail. */
	private final String LEVEL_OF_DETAIL = "LOD";
	
	/** The local name of the exterior surface area tag. */
	private final String EXTERIOR_SURFACE_AREA = "exteriorSurfaceArea";
	
	/** The local name of the shared surface area. */
	private final String SHARED_SURFACE_AREA = "sharedSurfaceArea";
	
	/** The local name of the cold surface area above ground. */
	private final String THERMAL_BOUNDARY_AREA_ABOVE_GROUND = "thermalBoundaryAreaAboveGround";
	
	/** The local name of the cold surface area below ground. */
	private final String THERMAL_BOUNDARY_AREA_BELOW_GROUND = "thermalBoundaryAreaBelowGround";
	
	/** The local name of the azimuth. */
	private final String AZIMUTH = "azimuth";
	
	/** The local name of the inclination. */
	private final String INCLINATION = "inclination";
	
	/** The short wave irradiance. */
	private final String SHORT_WAVE_IRRADIANCE = "shortWaveIrradiance";
	
	/** The shell properties. */
	private final String SHELL_PROPERTIES = "shellProperties";
	
	/**
	 * @return the namespace URI of the Energy ADE.
	 */
	@Override
	public String handlesNamespaceURI() {
		return NAMESPACE_URI;
	}
	
	/**
	 * @return the package of the Energy ADE.
	 */
	@Override
	public String getADEPackage() {
		return ADE_PACKAGE;
	}
	
	/**
	 * Looks into the extensions of the given building instance and marshals all found Energy ADE specific
	 * extensions. 
	 * 
	 * @param building The CityDoctor building that may contain Energy ADE specific extensions.
	 * @return a list of encapsulated Energy ADE specific extensions wrapped by ADEComponent instances.
	 */
	@Override
	public ArrayList<ADEComponent> marshal(de.hft.stuttgart.citydoctor.datastructure.Building building)
			throws JAXBException, SAXException {
		return marshalAbstractBuilding(building.getExtensions(NAMESPACE_URI));
	}

	/**
	 * Looks into the extensions of the given building part instance and marshals all found Energy ADE specific
	 * extensions. 
	 * 
	 * @param buildingPart The CityDoctor building part that may contain Energy ADE specific extensions.
	 * @return a list of encapsulated Energy ADE specific extensions wrapped by ADEComponent instances.
	 */
	@Override
	public ArrayList<ADEComponent> marshal(de.hft.stuttgart.citydoctor.datastructure.BuildingPart buildingPart)
			throws JAXBException, SAXException {
		return marshalAbstractBuilding(buildingPart.getExtensions(NAMESPACE_URI));
	}
	
	/**
	 * This method performs the actual job for its callers marshal(building) and marshal(buildingPart).
	 * 
	 * @param extensionList The list of the abstract building's extension instances which may contain Energy ADE
	 * specific extensions.
	 * @return a list of encapsulated Energy ADE specific extensions wrapped by ADEComponent instances.
	 */
	private ArrayList<ADEComponent> marshalAbstractBuilding(ArrayList<Extension> extensionList) throws JAXBException {
		ArrayList<ADEComponent> adeComponentList = new ArrayList<ADEComponent>();
		for (Extension extension : extensionList) {
			BuildingUnit energyADEBuilding = (BuildingUnit) extension;
			JAXBMarshaller marshaller = getJAXBMarshaller();
			ObjectFactory objectFactory = new ObjectFactory();
			
			// Process the level of detail
			LOD lod = energyADEBuilding.getLOD();
			JAXBElement<LOD> jaxbLODElement = objectFactory.createLOD(lod);
			Element lodElement = marshaller.marshalDOMElement(jaxbLODElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(lodElement));
			
			// Process the building physics
			PhysicsProperties physicsProperties = energyADEBuilding.getPhysicsProperties();
			JAXBElement<PhysicsProperties> jaxbBuildingPhysicsElement =
					objectFactory.createPhysicsProperties(physicsProperties);
			Element physicsPropertiesElement = 
					marshaller.marshalDOMElement(jaxbBuildingPhysicsElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(physicsPropertiesElement));
			
			// Process spatial properties
			SpatialProperties spacialProperties = energyADEBuilding.getSpatialProperties();
			JAXBElement<SpatialProperties> jaxbSpacialPropertiesElement = 
					objectFactory.createSpatialProperties(spacialProperties);
			Element buildingDimensionsElement = 
					marshaller.marshalDOMElement(jaxbSpacialPropertiesElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(buildingDimensionsElement));
			
			// Process energy usages
			HashMap<EndUseType, EndUse> endUseMap = energyADEBuilding.getEndUseMap();
			for (EndUseType endUseType : endUseMap.keySet()) {
				EndUse endUse = endUseMap.get(endUseType);
				JAXBElement<EndUse> jaxbEndUseElement =	objectFactory.createEndUse(endUse);
				Element endUseElement =	marshaller.marshalDOMElement(jaxbEndUseElement, getJAXBContext());
				adeComponentList.add(new ADEComponent(endUseElement));
			}
			
			// Process usage zones
			UsageZoneList usageZoneList = energyADEBuilding.getUsageZoneList();
			JAXBElement<UsageZoneList> jaxbUsageZoneListElement =
					objectFactory.createUsageZoneList(usageZoneList);
			Element usageZoneListElement =
					marshaller.marshalDOMElement(jaxbUsageZoneListElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(usageZoneListElement));
		}
		return adeComponentList;
	}

	/**
	 * Looks into the extensions of the given boundary surface instance and marshals all found Energy ADE
	 * specific extensions. 
	 * 
	 * @param boundarySurface The CityDoctor boundary surface that may contain Energy ADE specific extensions.
	 * @return a list of encapsulated Energy ADE specific extensions wrapped by ADEComponent instances.
	 */
	@Override
	public ArrayList<ADEComponent> marshal(BoundarySurface boundarySurface)
			throws JAXBException, SAXException {
		ArrayList<ADEComponent> adeComponentList = new ArrayList<ADEComponent>();
		ArrayList<Extension> extensionList = boundarySurface.getExtensions(NAMESPACE_URI);
		for (Extension extension : extensionList) {
			BoundarySurfaceUnit boundarySurfaceUnit = (BoundarySurfaceUnit) extension;
			JAXBMarshaller marshaller = getJAXBMarshaller();
			ObjectFactory objectFactory = new ObjectFactory();
			
			// Marshal the exterior surface area
			JAXBElement<Double> jaxbExteriorSurfaceAreaElement =
					objectFactory.createExteriorSurfaceArea(boundarySurfaceUnit.getExteriorSurfaceArea());
			Element exteriorSurfaceAreaElement =
					marshaller.marshalDOMElement(jaxbExteriorSurfaceAreaElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(exteriorSurfaceAreaElement));
			
			// Marshal the shared surface area
			JAXBElement<Double> jaxbSharedSurfaceAreaElement =
					objectFactory.createSharedSurfaceArea(boundarySurfaceUnit.getSharedSurfaceArea());
			Element sharedSurfaceAreaElement =
					marshaller.marshalDOMElement(jaxbSharedSurfaceAreaElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(sharedSurfaceAreaElement));
			
			// Marshal the cold surface area above ground
			JAXBElement<Double> jaxbThermalBoundaryAreaAboveGroundElement =
					objectFactory.createThermalBoundaryAreaAboveGround(boundarySurfaceUnit.getThermalBoundaryAreaAboveGround());
			Element thermalBoundaryAreaAboveGroundElement =
					marshaller.marshalDOMElement(jaxbThermalBoundaryAreaAboveGroundElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(thermalBoundaryAreaAboveGroundElement));
			
			// Marshal the cold surface area below ground
			JAXBElement<Double> jaxbThermalBoundaryAreaBelowGroundElement =
					objectFactory.createThermalBoundaryAreaBelowGround(boundarySurfaceUnit.getThermalBoundaryAreaBelowGround());
			Element thermalBoundaryAreaBelowGroundElement =
					marshaller.marshalDOMElement(jaxbThermalBoundaryAreaBelowGroundElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(thermalBoundaryAreaBelowGroundElement));
			
			// Marshal the azimuth
			JAXBElement<Integer> jaxbAzimuthElement = objectFactory.createAzimuth(boundarySurfaceUnit.getAzimuth());
			Element azimuthElement = marshaller.marshalDOMElement(jaxbAzimuthElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(azimuthElement));
			
			// Marshal the inclination
			JAXBElement<Integer> jaxbInclinationElement = 
					objectFactory.createInclination(boundarySurfaceUnit.getInclination());
			Element inclinationElement = marshaller.marshalDOMElement(jaxbInclinationElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(inclinationElement));
			
			// Marshal the short wave irradiance
			JAXBElement<List<Integer>> jaxbShortWaveIrradianceElement = 
					objectFactory.createShortWaveIrradiance(boundarySurfaceUnit.getShortWaveIrradiance());
			Element shortWaveIrradianceElement =
					marshaller.marshalDOMElement(jaxbShortWaveIrradianceElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(shortWaveIrradianceElement));
			
			// Marshal the shell properties
			JAXBElement<ShellProperties> jaxbShellPropertiesElement =
					objectFactory.createShellProperties(boundarySurfaceUnit.getShellProperties());
			Element shellPropertiesElement =
					marshaller.marshalDOMElement(jaxbShellPropertiesElement, getJAXBContext());
			adeComponentList.add(new ADEComponent(shellPropertiesElement));
		}
		return adeComponentList;
	}
	
	/**
	 * Unmarshals all Energy ADE components of a given CityGML building instance. Energy ADE specific
	 * components will be wrapped in instances of ADEComponent.
	 * 
	 * @param the CityGML building that may contain some Energy ADE specific extensions.
	 * @return the encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the abstract building.
	 */
	@Override
	public ArrayList<Extension>	unmarshal(Building building) throws JAXBException, SAXException {
		return unmarshalAbstractBuilding(building.getGenericApplicationPropertyOfAbstractBuilding());
	}
	
	/**
	 * Unmarshals all Energy ADE components of a given CityGML buildingPart instance. Energy ADE specific
	 * components will be wrapped in instances of ADEComponent.
	 * 
	 * @param the CityGML building part that may contain some Energy ADE specific extensions.
	 * @return the encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the abstract building.
	 */
	@Override
	public ArrayList<Extension> unmarshal(BuildingPart buildingPart)
			throws JAXBException, SAXException {
		return unmarshalAbstractBuilding(buildingPart.getGenericApplicationPropertyOfAbstractBuilding());
	}
	
	/**
	 * Performs the actual job for its callers. Please search for them for more details.
	 * 
	 * @param A list of ADEComponent instances which may contain some Energy ADE specific extensions.
	 * @return The encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the abstract building.
	 */
	private ArrayList<Extension> unmarshalAbstractBuilding(List<ADEComponent> adeComponentList)
			throws JAXBException {
		ArrayList<Extension> extensionList = new ArrayList<>();
		BuildingUnit buildingUnit = new BuildingUnit();
		Unmarshaller jaxbUnmarshaller = getJAXBUnmarshaller();
		for (ADEComponent adeComponent : adeComponentList) {
			if (adeComponent.getNamespaceURI().equals(NAMESPACE_URI)) {
				extensionList.add(buildingUnit);
				JAXBElement<?> element = (JAXBElement<?>) jaxbUnmarshaller.unmarshal(adeComponent.getContent());
				if (adeComponent.getLocalName().equals(PHYSICS_PROPERTIES)) {
					buildingUnit.setPhysicsProperties((PhysicsProperties) element.getValue());
				} else if (adeComponent.getLocalName().equals(SPATIAL_PROPERTIES)) {
					buildingUnit.setSpatialProperties((SpatialProperties) element.getValue());
				} else if (adeComponent.getLocalName().equals(END_USE)) {
					EndUse endUse = (EndUse) element.getValue();
					buildingUnit.getEndUseMap().put(endUse.getType(), endUse);
				} else if (adeComponent.getLocalName().equals(USAGE_ZONE_LIST)) {
					buildingUnit.setUsageZoneList((UsageZoneList) element.getValue());
				} else if (adeComponent.getLocalName().equals(LEVEL_OF_DETAIL)) {
					buildingUnit.setLOD((LOD) element.getValue());
				}
			}
		}
		return extensionList;
	}

	/**
	 * Unmarshals all Energy ADE components of a given CityGML roof surface instance. Energy ADE specific
	 * components will be wrapped in instances of ADEComponent.
	 * 
	 * @param the CityGML roof surface that may contain some Energy ADE specific extensions.
	 * @return the encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the roof surface.
	 */
	@Override
	public ArrayList<Extension> unmarshal(RoofSurface boundarySurface)
			throws JAXBException, SAXException {
		return unmarshalBoundarySurface(boundarySurface.getGenericApplicationPropertyOfBoundarySurface());
	}

	/**
	 * Unmarshals all Energy ADE components of a given CityGML wall surface instance. Energy ADE specific
	 * components will be wrapped in instances of ADEComponent.
	 * 
	 * @param the CityGML wall surface that may contain some Energy ADE specific extensions.
	 * @return the encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the wall surface.
	 */
	@Override
	public ArrayList<Extension> unmarshal(WallSurface boundarySurface)
			throws JAXBException, SAXException {
		return unmarshalBoundarySurface(boundarySurface.getGenericApplicationPropertyOfBoundarySurface());
	}

	/**
	 * Unmarshals all Energy ADE components of a given CityGML ground surface instance. Energy ADE specific
	 * components will be wrapped in instances of ADEComponent.
	 * 
	 * @param the CityGML ground surface that may contain some Energy ADE specific extensions.
	 * @return the encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the ground surface.
	 */
	@Override
	public ArrayList<Extension> unmarshal(GroundSurface boundarySurface)
			throws JAXBException, SAXException {
		return unmarshalBoundarySurface(boundarySurface.getGenericApplicationPropertyOfBoundarySurface());
	}

	/**
	 * Unmarshals all Energy ADE components of a given CityGML closure surface instance. Energy ADE specific
	 * components will be wrapped in instances of ADEComponent.
	 * 
	 * @param the CityGML closure surface that may contain some Energy ADE specific extensions.
	 * @return the encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the closure surface.
	 */
	@Override
	public ArrayList<Extension> unmarshal(ClosureSurface boundarySurface)
			throws JAXBException, SAXException {
		return unmarshalBoundarySurface(boundarySurface.getGenericApplicationPropertyOfBoundarySurface());
	}

	/**
	 * Unmarshals all Energy ADE components of a given CityGML outer floor surface instance. Energy ADE specific
	 * components will be wrapped in instances of ADEComponent.
	 * 
	 * @param the CityGML outer floor surface that may contain some Energy ADE specific extensions.
	 * @return the encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the outer floor surface.
	 */
	@Override
	public ArrayList<Extension> unmarshal(OuterFloorSurface boundarySurface)
			throws JAXBException, SAXException {
		return unmarshalBoundarySurface(boundarySurface.getGenericApplicationPropertyOfBoundarySurface());
	}

	/**
	 * Unmarshals all Energy ADE components of a given CityGML outer ceiling surface instance. Energy ADE specific
	 * components will be wrapped in instances of ADEComponent.
	 * 
	 * @param the CityGML outer ceiling surface that may contain some Energy ADE specific extensions.
	 * @return the encapsulated Energy ADE specific extensions. Returns null, if no extension is defined for
	 * the outer ceiling surface.
	 */
	@Override
	public ArrayList<Extension> unmarshal(OuterCeilingSurface boundarySurface)
			throws JAXBException, SAXException {
		return unmarshalBoundarySurface(boundarySurface.getGenericApplicationPropertyOfBoundarySurface());
	}
	
	/**
	 * Performs the actual job for its callers. Please search for them in order to get more details.
	 * 
	 * @param The list of ADEComponent instances that may contain some Energy ADE specific extensions.
	 * @return The encapsulated Energy ADE specific extensions. Returns null, if no extension are found.
	 */
	private ArrayList<Extension> unmarshalBoundarySurface(List<ADEComponent> adeComponentList)
			throws JAXBException, SAXException {
		ArrayList<Extension> extensionList = new ArrayList<>();
		BoundarySurfaceUnit boundarySurfaceUnit = new BoundarySurfaceUnit();
		Unmarshaller jaxbUnmarshaller = getJAXBUnmarshaller();
		for (ADEComponent adeComponent : adeComponentList) {
			if (adeComponent.getNamespaceURI().equals(NAMESPACE_URI)) {
				extensionList.add(boundarySurfaceUnit);
				JAXBElement<?> element = (JAXBElement<?>) jaxbUnmarshaller.unmarshal(adeComponent.getContent());
				if (adeComponent.getLocalName().equals(EXTERIOR_SURFACE_AREA)) {
					boundarySurfaceUnit.setExteriorSurfaceArea((Double) element.getValue());
				} else if (adeComponent.getLocalName().equals(SHARED_SURFACE_AREA)) {
					boundarySurfaceUnit.setSharedSurfaceArea((Double) element.getValue());
				} else if (adeComponent.getLocalName().equals(THERMAL_BOUNDARY_AREA_ABOVE_GROUND)) {
					boundarySurfaceUnit.setThermalBoundaryAreaAboveGround((Double) element.getValue());
				} else if (adeComponent.getLocalName().equals(THERMAL_BOUNDARY_AREA_BELOW_GROUND)) {
					boundarySurfaceUnit.setThermalBoundaryAreaBelowGround((Double) element.getValue());
				} else if (adeComponent.getLocalName().equals(AZIMUTH)) {
					boundarySurfaceUnit.setAzimuth((Integer) element.getValue());
				} else if (adeComponent.getLocalName().equals(INCLINATION)) {
					boundarySurfaceUnit.setInclination((Integer) element.getValue());
				} else if (adeComponent.getLocalName().equals(SHORT_WAVE_IRRADIANCE)) {
					boundarySurfaceUnit.setShortWaveIrradiance((List<Integer>) element.getValue());
				} else if (adeComponent.getLocalName().equals(SHELL_PROPERTIES)) {
					boundarySurfaceUnit.setShellProperties((ShellProperties) element.getValue());
				}
			}
		}
		return extensionList;			
	}

}
